#!/usr/bin/python

# Real-time sync methods. Builds and extends upon those available in sync.syncactions

import posixpath, os, logging, datetime, time
import sync


def localFileMod(file_localpath, db_engine, db_session, 
		sync_collection, new_localfile = True):
	'''Check whether the resource has been modified from a previous state'''

	# Check filename for illegal characters
	if not sync.checkFileName(file_localpath):
		return sync_collection['Collection-Name'], sync.LOCAL_ILLEGAL_FILENAME, file_localpath

	# Check file metadata 
	if os.path.isfile(file_localpath):

		# Retrieve file metadata and database information
		file_cdate, file_mdate = sync.getFileDates(file_localpath)
		try: md5_beg, md5_end = sync.checkMd5(file_localpath, sync.FILE_CHUNK)
		except: return sync_collection['Collection-Name'], sync.RT_FAIL, file_localpath
		
		# Retrieve the database for the local file
		try: dbfile_ = sync.getLocalFileRecord(file_localpath, db_session)
		except: return sync_collection['Collection-Name'], sync.RT_FAIL, file_localpath

		# New files not yet in the database
		if (dbfile_ == None):
			return sync_collection['Collection-Name'], sync.RT_UPLOAD_NEW, file_localpath

		# Existing files
		elif (dbfile_ is not None):

			# Check to see if the file has changed
			if (file_mdate > dbfile_.mdate) | (file_cdate > dbfile_.cdate):

				# Check to see if the md5 hashes differ
				if (md5_beg != dbfile_.md5_beg) | (md5_end != dbfile_.md5_end):
					return sync_collection['Collection-Name'], sync.RT_UPLOAD_UPDATE, \
						file_localpath

	# No modifications or unable to find file or database record, take no action
	return sync_collection['Collection-Name'], sync.RT_NOTNEEDED, file_localpath

def checkLocalFileDelete(file_localpath, db_engine, db_session, sync_collection):
	'''Verify that a local file has been deleted before removing it from the sync
		database or from the online collection'''

	# Allow time to go by for file operations to finish
	time.sleep(sync.OPS_PAUSE)

	# Check to see if the file has been removed
	if not os.path.exists(file_localpath):
		return sync_collection['Collection-Name'], sync.RT_DELETE_REMOTE, file_localpath

	# Check to see if file deleted as part of save operation
	elif os.path.exists(file_localpath):
		localFileMod(file_localpath, db_engine, db_session, sync_collection)

	else: return sync_collection['Collection-Name'], sync.RT_NOTNEEDED, file_localpath


def localFolderChanges(file_localpath, db_engine, db_session, sync_collection):
	'''Analyze the contents of a local folder on disk and determine what has changed'''

	if os.path.exists(file_localpath):
		logging.info('Checking for local changes in: ' + str(file_localpath))
		try: 
			folder_changes = sync.localFolderChanges(sync_collection, file_localpath, db_session)
			return (sync_collection['Collection-Name'], sync.RT_SUCCESS, file_localpath), folder_changes
		except: pass

	return (sync_collection['Collection-Name'], sync.RT_FAIL, file_localpath), []


# Real-Time Sync Operations

def remoteFolderChanges(sync_collection, sync_backend, db_engine, db_session):
	'''Check a remote collection for changes'''
	
	rd_ = posixpath.join(sync_backend.server_url, sync_collection['Remote-Path'])
	
	try: 
		remote_changes = sync_backend.remoteFolderChanges(sync_collection,
			rd_, sync_collection['Local-Path'], db_session)
		return (sync_collection['Collection-Name'], sync.RT_SUCCESS, rd_), remote_changes
	
	except: return (sync_collection['Collection-Name'], sync.RT_FAIL, rd_), []


def deleteRemoteFile(file_localpath, db_engine, db_session,
		sync_collection, sync_backend):
	'''Delete a file from the remote collection'''

	file_relpath = sync.abs2rel(file_localpath, sync_collection['Local-Path'])
	file_remotepath = posixpath.join(sync_backend.server_url,
		sync_collection['Remote-Path'], file_relpath)

	# Check to make sure that the file has been deleted
	if not os.path.exists(file_localpath):

		# Retrieve dbfile and remote reference
		dbfile_ = sync.getLocalFileRecord(file_localpath, db_session)
		try: rdfile_ = sync_backend.remoteFileDetails(file_remotepath)
		except sync.errors.ConnectionError, e:
			return sync_collection['Collection-Name'], \
				sync.RT_FAIL, file_localpath
		
		# If there is a database entry and remote entry for the file,
		# run the remote delete operation
		if (rdfile_ is not None) & (dbfile_ is not None):
			result_ = sync.deleteRemoteFile(rdfile_, dbfile_, db_session, sync_backend)

		if result_: return sync_collection['Collection-Name'], \
			sync.RT_SUCCESS, file_localpath
		else: return sync_collection['Collection-Name'], \
			sync.RT_FAIL, file_localpath
	
	# If the file still exists on disk, do nothing
	else: return sync_collection['Collection-Name'], \
		sync.RT_NOTNEEDED, file_localpath


def downloadFile(file_localpath, db_engine, db_session,
		sync_collection, sync_backend):
	'''Download a new or updated file from the server'''

	if not sync.checkFileName(file_localpath):
		return sync_collection['Collection-Name'], sync.LOCAL_ILLEGAL_FILENAME, file_localpath

	file_relpath = sync.abs2rel(file_localpath, sync_collection['Local-Path'])
	file_remotepath = posixpath.join(sync_backend.server_url,
		sync_collection['Remote-Path'], file_relpath)
	rdfile_ = sync_backend.remoteFileDetails(file_remotepath)
	rdfile_.local_path = file_localpath

	# Check if there is a local copy of the file
	if not os.path.exists(file_localpath):

		# Check to see if there is a record in the database
		dbfile_ = sync.getLocalFileRecord(file_localpath, db_session)

		# New File - Download to the local disk
		if dbfile_ is None:
			if (rdfile_.md5_beg is None) & (rdfile_.md5_end is None):
			 	result_ = sync.downloadNewFile(rdfile_, db_session, sync_backend, server_file=True)
			else: result_ = sync.downloadNewFile(rdfile_, db_session, sync_backend)

		elif dbfile_ is not None:

			# Remote and local file have the same etag or md5 signature
			if ((rdfile_.file_etag is not None) & (dbfile_.file_etag == rdfile_.file_etag)) \
				| ((rdfile_.md5_beg is not None) & (dbfile_.md5_beg == rdfile_.md5_beg)\
					& (dbfile_.md5_end == rdfile_.md5_end)):
					return sync_collection['Collection-Name'], sync.RT_NOTNEEDED, file_localpath

			# Remote and local file have different etag or md5 signatures
			elif ((rdfile_.file_etag is not None) & (dbfile_.file_etag != rdfile_.file_etag)) \
				& ((rdfile_.md5_beg != dbfile_.md5_beg) | (rdfile_.md5_end != dbfile_.md5_end)):

				# Check mdate
				if (rdfile_.mdate > dbfile_.mdate):
					result_, conflict_filepath = sync.replaceLocalFile(dbfile_, rdfile_,
						db_session, sync_backend)

		if result_:
			return sync_collection['Collection-Name'], sync.RT_SUCCESS, file_localpath

	return sync_collection['Collection-Name'], sync.RT_FAIL, file_localpath


def uploadFile(file_localpath, db_engine, db_session, 
		sync_collection, sync_backend, new_localfile = True):
	'''Upload a new or updated file to the server'''

	file_relpath = sync.abs2rel(file_localpath, sync_collection['Local-Path'])
	file_remotepath = posixpath.join(sync_backend.server_url,
		sync_collection['Remote-Path'], file_relpath)

	# Retrieve File Information
	file_cdate, file_mdate = sync.getFileDates(file_localpath)
	md5_beg, md5_end = sync.checkMd5(file_localpath, sync.FILE_CHUNK)

	# Check to see if the upload destination exists
	folder_check = sync_backend.checkFolder(
		posixpath.dirname(file_remotepath), create_folder=True)

	if (folder_check == True) & (db_session is not None):
		result_ = None

		# Retrieve file information and remote file information
		dbfile_ = sync.getLocalFileRecord(file_localpath, db_session)

		logging.info('Requesting remote object for ' + file_localpath)

		# Retrieve remote file reference
		try: 
			rdfile_ = sync_backend.remoteFileDetails(file_remotepath)
		except sync.errors.ConnectionError, e:
			return sync_collection['Collection-Name'], sync.RT_FAIL, file_localpath
		except sync.errors.SyncError, e:
			if e.type == sync.errors.ERROR_NOTFOUND: rdfile_ = None

		logging.info('Retrieved remote file object: ' + file_localpath)

		# New file on disk
		if ((new_localfile == True) & (rdfile_ is None)) \
			| ((rdfile_ is None) & (dbfile_ is None)):
			try:
				result_ = sync.uploadNewFile(
					file_localpath, file_remotepath, # Local and remote paths
					(file_cdate, file_mdate), # File cdate and mdate
					(md5_beg, md5_end),
					db_session, sync_backend)
			except: result_ = False

		# Update existing file
		elif (dbfile_ is not None) & (rdfile_ is not None):

			# Check for remote differences
			if (md5_beg != rdfile_.md5_beg) | (md5_end != rdfile_.md5_end):

				# Local file newer than remote
				if file_mdate > rdfile_.mdate:
					logging.info('Uploading ' + str(file_localpath))
					try: 
						result_ = sync.uploadUpdatedFile(dbfile_, file_remotepath,
							(file_cdate, file_mdate),
							(md5_beg, md5_end),
							db_session, sync_backend)
					except: result_ = False

				else: 
					return sync_collection['Collection-Name'], \
						sync.RT_NOTNEEDED, file_localpath

		# Remote file is missing
		elif (dbfile_ is not None) & (rdfile_ is None):
			try: 
				result_ = sync.uploadUpdatedFile(dbfile_, file_remotepath,
					(file_cdate, file_mdate),
					(md5_beg, md5_end),
					db_session, sync_backend)
			except: result_ = False

		# Record of upload missing
		elif (dbfile_ is None) & (rdfile_ is not None):

			# Check against local file, add to database
			result_, conflict_filepath = sync.mergeExistingFile(
				file_localpath, # Local path
				file_remotepath, # remote path of file,
				(file_cdate, file_mdate), # File cdate and mdate
				rdfile_, # Remote file object,
				db_session, sync_backend)

		if result_: return sync_collection['Collection-Name'], \
			sync.RT_SUCCESS, file_localpath

	# Unable to complete transfer
	return sync_collection['Collection-Name'], sync.RT_FAIL, file_localpath